# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import asyncio
import inspect
import json
import re
import time
from asyncio import Task, run

from kubernetes import config
from kubernetes.client import ApiException, CoreV1Api, V1EnvVar, V1ObjectMeta, V1Pod
from traitlets import Callable, Dict, Integer, List, Unicode
from traitlets.config import LoggingConfigurable
from urllib3.exceptions import MaxRetryError

from grader_service.autograding.kube.util import get_current_namespace, make_pod
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.orm import Assignment, Lecture, Submission
from grader_service.orm.assignment import json_serial


class GraderPod(LoggingConfigurable):
    """
    Wrapper for a kubernetes pod that supports polling of the pod's status.
    """

    poll_interval = Integer(
        default_value=5, allow_none=False, help="Time in sec to wait before status is polled again."
    ).tag(config=True)
    max_grading_time = Integer(
        default_value=3600,
        allow_none=False,
        help="Maximal grading time in sec before the grading fails",
    ).tag(config=True)

    def __init__(self, pod: V1Pod, api: CoreV1Api, **kwargs):
        super().__init__(**kwargs)
        self.pod = pod
        self._client = api
        # Ensure the event loop exists or create a new one
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            # Create a new event loop if none exists
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self._polling_task = self.loop.create_task(self._poll_status())

    def stop_polling(self) -> None:
        self._polling_task.cancel()

    def poll(self) -> str:
        return self.loop.run_until_complete(self.polling)

    @property
    def polling(self) -> Task:
        return self._polling_task

    @property
    def name(self) -> str:
        return self.pod.metadata.name

    @property
    def namespace(self) -> str:
        return self.pod.metadata.namespace

    # Watch for pod status changes instead of polling in intervals.
    async def _poll_status(self) -> str:
        meta: V1ObjectMeta = self.pod.metadata
        start_time = time.time()
        interval = self.poll_interval
        timeout = self.max_grading_time

        while True:
            # Read the pod status using read_namespaced_pod_status
            pod_status = self._client.read_namespaced_pod_status(
                name=meta.name, namespace=meta.namespace
            )

            # Extract the pod phase (status)
            phase = pod_status.status.phase
            self.log.debug(f"Pod {meta.name} is currently in {phase} phase.")

            # Exit conditions
            if phase in ["Succeeded", "Failed"]:  # Stop polling if pod completes or fails
                self.log.info(f"Pod {meta.name} has finished with phase: {phase}.")
                return phase

            # Check if timeout is reached
            elapsed_time = time.time() - start_time
            if elapsed_time > timeout:
                print(f"Polling timed out after {timeout} seconds.")
                return "Failed"

            # Wait for the next poll
            time.sleep(interval)


def _get_image_name(lecture: Lecture, assignment: Assignment = None) -> str:
    """
    Default implementation of the resolve_image_name method
    which return the lecture code followed by '_image'.
    All the functions have the lecture and assignment available as parameters.
    The function can either be sync or async.
    :param lecture: Lecture to build the image name.
    :param assignment: Assignment to build the image name.
    :return: The image name as a string.
    """
    return f"{lecture.code}_image"


class KubeAutogradeExecutor(LocalAutogradeExecutor):
    """
    Runs an autograde job in a kubernetes cluster as a pod.
    The cluster has to have a shared persistent
    volume claim that is mounted in the input
    and output directories so that both the service and
    the executor pods have access to the files.
    The service account of the grader service has to have
    permission to get, update, create and delete pods, pod status and pod logs.
    """

    # Annotations to attach to the Kubernetes pod
    annotations = Dict(
        default_value={},
        allow_none=True,
        help="Annotations to associate with the pod. Defaults to an empty dictionary.",
    ).tag(config=True)

    # Name of the executable for converting the grader
    convert_executable = Unicode(
        "grader-convert",
        allow_none=False,
        help="The executable name for the grader container conversion. Defaults to 'grader-convert'.",
    ).tag(config=True)

    # List of extra volumes to attach to the pod
    extra_volumes = List(
        default_value=[],
        allow_none=False,
        help="List of extra volumes to attach to the pod. Defaults to an empty list.",
    ).tag(config=True)

    # List of extra volume mounts to attach to the container
    extra_volume_mounts = List(
        default_value=[],
        allow_none=False,
        help="List of extra volume mounts for the container. Defaults to an empty list.",
    ).tag(config=True)

    # Configuration for the image name, with a callable for resolution
    image_config_path = Unicode(
        default_value=None,
        allow_none=True,
        help="Deprecated in v0.7, will be removed in v0.8, use pre_spawn_hook to dynamically change the autograding pod image.",
    ).tag(config=True)

    # Image pull policy for the Kubernetes pod
    image_pull_policy = Unicode(
        default_value="Always",
        allow_none=False,
        help="The image pull policy for the pod. Defaults to 'Always'.",
    ).tag(config=True)

    # Dictionary to store image pull secrets, helpful when pulling from private registries
    image_pull_secrets = List(
        default_value=[],
        help="""Autograding pod image pull secrets list (str). 
                                      Used for pulling images from private registries. Defaults to None.""",
        key_trait=Unicode(),
        value_trait=Unicode(),
        allow_none=True,
    ).tag(config=True)

    # Kubernetes context to load configuration from (in-cluster if None)
    kube_context = Unicode(
        default_value=None,
        allow_none=True,
        help="Kubernetes context to load the config from. If None, the in-cluster config is used.",
    ).tag(config=True)

    # Labels to attach to the Kubernetes pod
    labels = Dict(
        default_value={},
        allow_none=True,
        help="Labels to associate with the pod. Defaults to an empty dictionary.",
    ).tag(config=True)

    # Namespace where grader pods will be deployed
    namespace = Unicode(
        default_value=None,
        allow_none=True,
        help="Namespace for deploying grader pods. If None, the current namespace will be used. "
        "If changed, roles for ServiceAccount need to be applied.",
    ).tag(config=True)

    # Pre-spawn Hook configuration
    pre_spawn_hook = Callable(
        default_value=None,
        allow_none=True,
        help="""
        A callable function that executes before the autograding pod spawns. 
        Use this hook to customize pod configurations based on the specific lecture and assignment.

        **Parameters:**
        - `lecture` (Lecture): The lecture associated with the submission.
        - `assignment` (Assignment): The assignment associated with the submission.
        - `autograding_pod` (V1Pod): The current Kubernetes pod object for autograding.
        **Returns:**
        - V1Pod spec for autograding pod
        **Purpose:**
        This hook allows dynamic modification of pod specifications, such as:
        - Updating selectors.
        - Changing the container image.
        - Adjusting resource requests and limits.
        
        **Default:**
        None (no modifications will be applied unless a callable is provided).
        """,
    ).tag(config=True)

    # Callable to resolve the image name
    resolve_image_name = Callable(
        default_value=_get_image_name,
        allow_none=False,
        help="Deprecated in v0.7, will be removed in v0.8, use pre_spawn_hook to dynamically change the autograding pod image.",
    ).tag(config=True)

    # Callable for resolving node selector for a lecture
    resolve_node_selector = Callable(
        default_value=lambda _: None,
        allow_none=False,
        help="""Deprecated in v0.7, will be removed in v0.8, use pre_spawn_hook to dynamically set the node selectors of autograding pods.""",
    ).tag(config=True)

    # Tolerations for the autograding pod
    tolerations = Dict(
        default_value=None,
        help="""Autograding pod tolerations dictionary (str, str). 
                               Used to schedule pods on nodes with specific taints. Defaults to None.""",
        key_trait=Unicode(),
        value_trait=Unicode(),
        allow_none=True,
    ).tag(config=True)

    # User ID for the grader container (used to set user permissions inside the container)
    uid = Integer(
        default_value=1000,
        allow_none=False,
        help="User ID for the grader container. Defaults to 1000.",
    ).tag(config=True)

    # Dictionary for additional volume configuration
    volume = Dict(
        default_value={},
        allow_none=False,
        help="Dictionary for volume configuration. Defaults to an empty dictionary.",
    ).tag(config=True)

    def __init__(self, grader_service_dir: str, submission: Submission, **kwargs):
        super().__init__(grader_service_dir, submission, **kwargs)
        self.lecture = self.assignment.lecture
        if self.kube_context is None:
            self.log.info(
                f"Loading in-cluster config for kube executor of submission {self.submission.id}"
            )
            config.load_incluster_config()
        else:
            self.log.info(
                f"Loading cluster config '{self.kube_context}' "
                f"for kube executor of submission {self.submission.id}"
            )
            config.load_kube_config(context=self.kube_context)
        self.client = CoreV1Api()
        if self.namespace is None:
            self.log.info(f"Setting Namespace for submission {self.submission.id}")
            self.namespace = get_current_namespace()

    def _get_image(self) -> str:
        """
        Returns the image name based on the lecture and assignment.
        If an image config file exists and has
        been specified it will first be queried for an image name.
        If the image name cannot be found in the
        config file or none has been specified
        the image name will be determined by the resolve_image_name function
        which takes the lecture
        and assignment as parameters and is specified in the config.
        The default implementation of this function is to
         return the lecture code followed by '_image'.
        :return: The image name as determined by this method.
        """
        cfg = {}
        if self.image_config_path is not None:
            with open(self.image_config_path, "r") as f:
                cfg = json.load(f)
        try:
            lecture_cfg = cfg[self.lecture.code]
            if isinstance(lecture_cfg, str):
                return lecture_cfg
            else:
                return lecture_cfg[self.assignment.name]
        except KeyError:
            if inspect.iscoroutinefunction(self.resolve_image_name):
                return run(self.resolve_image_name(self.lecture, self.assignment))
            else:
                return self.resolve_image_name(self.lecture, self.assignment)

    def _get_autograde_pod_name(self) -> str:
        # sanitize username by converting to lowercase and replacing non-alphanumeric chars
        sanitized_username = re.sub(r"[^a-zA-Z0-9]+", "-", self.submission.user.name.lower())

        # truncate if too long to meet k8s pod name limits
        max_username_length = 50
        sanitized_username = sanitized_username[:max_username_length]

        # trim leading/trailing hyphens
        sanitized_username = sanitized_username.strip("-")

        return f"autograde-job-{sanitized_username}-{self.submission.id}"

    def _create_env(self) -> list[V1EnvVar]:
        env = [
            V1EnvVar(
                name="ASSIGNMENT_SETTINGS",
                value=json.dumps(self.assignment.settings.to_dict(), default=json_serial),
            )
        ]
        return env

    def _start_pod(self) -> GraderPod:
        """
        Starts a pod in the namespace
        with the commit hash as the name of the pod.
        The image is determined by the get_image method.
        :return:
        """
        # set standard config
        command = [
            self.convert_executable,
            "autograde",
            "-i",
            self.input_path,
            "-o",
            self.output_path,
            "-p",
            "*.ipynb",
            "--log-level=INFO",
            f"--ExecutePreprocessor.timeout={self.cell_timeout}",
        ]
        volumes = [self.volume] + self.extra_volumes
        volume_mounts = [
            {
                "name": "data",
                "mountPath": self.input_path,
                "subPath": self.relative_input_path + "/submission_" + str(self.submission.id),
            },
            {
                "name": "data",
                "mountPath": self.output_path,
                "subPath": self.relative_output_path + "/submission_" + str(self.submission.id),
            },
        ]
        volume_mounts = volume_mounts + self.extra_volume_mounts

        env = self._create_env()

        # create pod spec
        pod = make_pod(
            name=self._get_autograde_pod_name(),
            cmd=command,
            env=env,
            image=self._get_image(),
            image_pull_policy=self.image_pull_policy,
            image_pull_secrets=self.image_pull_secrets,
            working_dir="/",
            volumes=volumes,
            volume_mounts=volume_mounts,
            labels=self.labels,
            annotations=self.annotations,
            node_selector=self.resolve_node_selector(self.lecture),
            tolerations=self.tolerations,
            run_as_user=self.uid,
        )
        # run prespawn hook if it exists
        if callable(self.pre_spawn_hook):
            self.log.info(f"Running pre spawn hook for pod {pod.metadata.name}")
            pod = self.pre_spawn_hook(self.lecture, self.assignment, pod)

        # run grading pod
        self.log.info(f"Starting pod {pod.metadata.name} with command: {command}")
        pod = self.client.create_namespaced_pod(namespace=self.namespace, body=pod)
        # handle state of grading pod
        return GraderPod(pod, self.client, config=self.config)

    def _run(self):
        """
        Runs the autograding process in a kubernetes pod
        which has to have access to the files in the
        input and output directory through a persistent volume claim.
        :return: Coroutine
        """
        grader_pod = None
        try:
            grader_pod = self._start_pod()
            self.log.info(f"Started pod {grader_pod.name} in namespace {grader_pod.namespace}")
            status = grader_pod.poll()
            self.grading_logs = self._get_pod_logs(grader_pod)
            self.log.info("Pod logs:\n" + self.grading_logs)
            if status == "Succeeded":
                self.log.info(f"Pod {grader_pod.name} has successfully completed execution!")
            else:
                self.log.error(f"Pod {grader_pod.name} has failed execution!")
                self._delete_pod(grader_pod)
                raise RuntimeError(f"Pod {grader_pod.name} has failed execution!")
            # cleanup
            self._delete_pod(grader_pod)
        except ApiException as e:
            error_message = json.loads(e.body)
            if error_message["reason"] != "AlreadyExists" and grader_pod is not None:
                try:
                    self.client.delete_namespaced_pod(
                        name=grader_pod.name, namespace=grader_pod.namespace
                    )
                except ApiException:
                    pass
            self.log.error(f"{error_message['reason']}: {error_message['message']}")
            raise RuntimeError("Pod has failed execution!")
        except MaxRetryError:
            self.log.error(
                "Kubernetes client could not connect to cluster! "
                "Is it running and specified correctly?"
            )
            raise RuntimeError("Pod has failed execution!")

    def _delete_pod(self, pod: GraderPod):
        """
        Deletes the pod from the cluster after successful or failed execution.
        :param pod: The pod to delete.
        :return: None
        """
        self.log.info(
            f"Deleting pod '{pod.name}' in namespace '{pod.namespace}' "
            f"after execution status {pod.polling.result()}"
        )
        self.client.delete_namespaced_pod(name=pod.name, namespace=pod.namespace)

    def _get_pod_logs(self, pod: GraderPod) -> str:
        """
        Returns the logs of the pod that were output during execution.
        :param pod: The pod to retrieve the logs from.
        :return: The logs as a string.
        """
        api_response: str = self.client.read_namespaced_pod_log(
            name=pod.name, namespace=pod.namespace
        )
        return api_response.strip()



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class User(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, id=None, name=None, display_name=None):  # noqa: E501
        """User - a model defined in OpenAPI

        :param id: The id of this User.  # noqa: E501
        :type id: int
        :param name: The name of this User.  # noqa: E501
        :type name: str
        :param display_name: The display_name of this User.  # noqa: E501
        :type display_name: str
        """
        self.openapi_types = {
            'id': int,
            'name': str,
            'display_name': str
        }

        self.attribute_map = {
            'id': 'id',
            'name': 'name',
            'display_name': 'display_name'
        }

        self._id = id
        self._name = name
        self._display_name = display_name

    @classmethod
    def from_dict(cls, dikt) -> 'User':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The User of this User.  # noqa: E501
        :rtype: User
        """
        return util.deserialize_model(dikt, cls)

    @property
    def id(self) -> int:
        """Gets the id of this User.


        :return: The id of this User.
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id: int):
        """Sets the id of this User.


        :param id: The id of this User.
        :type id: int
        """

        self._id = id

    @property
    def name(self) -> str:
        """Gets the name of this User.


        :return: The name of this User.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name of this User.


        :param name: The name of this User.
        :type name: str
        """

        self._name = name

    @property
    def display_name(self) -> str:
        """Gets the display_name of this User.


        :return: The display_name of this User.
        :rtype: str
        """
        return self._display_name

    @display_name.setter
    def display_name(self, display_name: str):
        """Sets the display_name of this User.


        :param display_name: The display_name of this User.
        :type display_name: str
        """

        self._display_name = display_name



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum

from sqlalchemy import Column, Enum, ForeignKey, Integer
from sqlalchemy.orm import relationship

from grader_service.orm.base import Base, Serializable


class Scope(enum.IntEnum):
    student = 0
    tutor = 1
    instructor = 2
    admin = 3


class Role(Base, Serializable):
    __tablename__ = "takepart"
    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"), primary_key=True)
    lectid = Column(Integer, ForeignKey("lecture.id", ondelete="CASCADE"), primary_key=True)
    role = Column(Enum(Scope), nullable=False)

    lecture = relationship("Lecture", back_populates="roles")
    user = relationship("User", back_populates="roles")



import contextlib
import fnmatch
import glob
import hashlib
import io
import json
import logging
import os
import shutil
import stat
import sys
import tarfile
import traceback
import zipfile
from datetime import datetime
from logging import Logger
from typing import Any, Iterator, List, Optional, Tuple, Union

import dateutil.parser
from dateutil.tz import gettz
from nbformat.notebooknode import NotebookNode
from tornado.log import LogFormatter

from grader_service.api.models.assignment_settings import AssignmentSettings

# pwd is for unix passwords only, so we shouldn't import it on
# windows machines
if sys.platform != "win32":
    import pwd
else:
    pwd = None


def safe_unpack_archive(src, dest):
    if zipfile.is_zipfile(src):
        with zipfile.ZipFile(src, "r") as zip_ref:
            zip_ref.extractall(dest)
    elif tarfile.is_tarfile(src):
        with tarfile.open(src, "r:*") as tar_ref:
            tar_ref.extractall(dest)
    else:
        raise ValueError(f"Unsupported archive format: {src}")


def get_assignment_settings_from_env() -> AssignmentSettings:
    """Read env var and returns assignment settings

    Returns:
        AssignmentSettings: assignment settings
    """
    settings_json_str = os.getenv("ASSIGNMENT_SETTINGS", "{}")
    settings_dict = json.loads(settings_json_str)
    return AssignmentSettings.from_dict(settings_dict)


def is_task(cell: NotebookNode) -> bool:
    """Returns True if the cell is a task cell."""
    if "nbgrader" not in cell.metadata:
        return False
    return cell.metadata["nbgrader"].get("task", False)


def is_grade(cell: NotebookNode) -> bool:
    """Returns True if the cell is a grade cell."""
    if "nbgrader" not in cell.metadata:
        return False
    return cell.metadata["nbgrader"].get("grade", False)


def is_solution(cell: NotebookNode) -> bool:
    """Returns True if the cell is a solution cell."""
    if "nbgrader" not in cell.metadata:
        return False
    return cell.metadata["nbgrader"].get("solution", False)


def is_locked(cell: NotebookNode) -> bool:
    """Returns True if the cell source is locked (will be overwritten)."""
    if "nbgrader" not in cell.metadata:
        return False
    elif is_solution(cell):
        return False
    elif is_grade(cell):
        return True
    else:
        return cell.metadata["nbgrader"].get("locked", False)


def get_partial_grade(output, max_points, log=None):
    """
    Calculates partial grade for a cell, based on contents of
    output["data"]["text/plain"]. Returns a value between 0
    and max_points. Returns max_points (and a warning) for edge cases.
    """
    # check that output["data"]["text/plain"] exists
    if not output["data"]["text/plain"]:
        raise KeyError("output ['data']['text/plain'] does not exist")
    grade = output["data"]["text/plain"]
    # For partial credit, expecting grade to be a value between 0
    # and max_points
    # A valid value for key output["data"]["text/plain"] can be a
    # list or a string, so handle the string case
    if isinstance(grade, list):
        # grade is a list
        if len(grade) > 1:
            if log:
                warning_msg = """Cell output is {}, which is a list. For autograder tests, expecting output to indicate
                partial credit and be single value between 0.0 and max_points.
                Currently treating other output as full credit, but future
                releases may treat as error.""".format(grade)
                log.warning(warning_msg)
            return max_points
        # if a single value in list, set grade to that value
        grade = grade[0]
    try:
        # now that we have a single values for grade, can we
        # convert to a float?
        grade = float(grade)
    except ValueError:
        if log:
            warning_msg = """Cell output is {}, which cannot be converted to
            a float. For
            autograder tests, expecting output to indicate
            partial credit and be single value between 0.0 and max_points.
            Currently treating other output as full credit, but future releases
            may treat as error.""".format(grade)
            log.warning(warning_msg)
        return max_points
    if grade >= 0.0:
        if grade > max_points:
            raise ValueError("partial credit cannot be greater than maximum points for cell")
        return grade
    else:
        if log:
            warning_msg = """Cell output is {}, which is less than 0.0.
            This is strange.""".format(grade)
            log.warning(warning_msg)
        return max_points


def determine_grade(cell: NotebookNode, log: Logger = None) -> Tuple[Optional[float], float]:
    if not is_grade(cell):
        raise ValueError("cell is not a grade cell")

    max_points = float(cell.metadata["nbgrader"]["points"])
    if is_solution(cell):
        # if it's a solution cell and the checksum hasn't changed, that means
        # they didn't provide a response, so we can automatically give this a
        # zero grade
        if "checksum" in cell.metadata.nbgrader and cell.metadata.nbgrader[
            "checksum"
        ] == compute_checksum(cell):
            return 0, max_points
        else:
            return None, max_points

    elif cell.cell_type == "code":
        # for code cells, we look at the output. There are three options:
        # 1. output contains an error (no credit);
        # 2. output is a value greater than 0 (partial credit);
        # 3. output is something else, or nothing (full credit).
        for output in cell.outputs:
            # option 1: error, return 0
            if (
                output.output_type == "error"
                or output.output_type == "stream"
                and output.name == "stderr"
            ):
                return 0, max_points
            # if not error, then check for option 2, partial credit
            if output.output_type == "execute_result":
                # is there a single result that can be cast to a float?
                partial_grade = get_partial_grade(output, max_points, log)
                return partial_grade, max_points

        # otherwise, assume all fine and return all the points
        return max_points, max_points

    else:
        return None, max_points


def to_bytes(string: str) -> bytes:
    """A helper function for converting a string to bytes with utf-8 encoding."""
    return bytes(string.encode("utf-8"))


def compute_checksum(cell: NotebookNode) -> str:
    m = hashlib.md5()
    # add the cell source and type
    m.update(to_bytes(cell.source))
    m.update(to_bytes(cell.cell_type))

    # add whether it's a grade cell and/or solution cell
    m.update(to_bytes(str(is_grade(cell))))
    m.update(to_bytes(str(is_solution(cell))))
    m.update(to_bytes(str(is_locked(cell))))

    # include the cell id
    m.update(to_bytes(cell.metadata.nbgrader["grade_id"]))

    # include the number of points that the cell is worth, if it is a grade cell
    if is_grade(cell):
        m.update(to_bytes(str(float(cell.metadata.nbgrader["points"]))))

    return m.hexdigest()


def parse_utc(ts: Union[datetime, str]) -> datetime:
    """Parses a timestamp into datetime format, converting it to UTC if necessary."""
    if ts is None:
        return None
    if isinstance(ts, str):
        parts = ts.split(" ")
        if len(parts) == 3:
            ts = " ".join(parts[:2] + ["TZ"])
            tz = parts[2]
            try:
                tz = int(tz)
            except ValueError:
                tz = dateutil.tz.gettz(tz)
            ts = dateutil.parser.parse(ts, tzinfos=dict(TZ=tz))
        else:
            ts = dateutil.parser.parse(ts)
    if ts.tzinfo is not None:
        ts = (ts - ts.utcoffset()).replace(tzinfo=None)
    return ts


def to_numeric_tz(timezone):
    """Converts a timezone to a format which can be read by parse_utc."""
    return as_timezone(datetime.now(timezone.utc), timezone).strftime("%z")


def as_timezone(ts, timezone):
    """Converts UTC timestamp ts to have timezone tz."""
    if not timezone:
        return ts
    tz = gettz(timezone)
    if tz:
        return (ts + tz.utcoffset(ts)).replace(tzinfo=tz)
    else:
        return ts


def check_mode(path, read=False, write=False, execute=False):
    """Can the current user can rwx the path."""
    mode = 0
    if read:
        mode |= os.R_OK
    if write:
        mode |= os.W_OK
    if execute:
        mode |= os.X_OK
    return os.access(path, mode)


def check_directory(path, read=False, write=False, execute=False):
    """Does that path exist and can the current user rwx."""
    if os.path.isdir(path) and check_mode(path, read=read, write=write, execute=execute):
        return True
    else:
        return False


def get_osusername():
    """Get the username of the current process."""
    if pwd is None:
        raise OSError("get_username cannot be called on Windows")
    return pwd.getpwuid(os.getuid())[0]


def get_username():
    """Get the username, use os username but override if username is jovyan ."""
    osname = get_osusername()
    if osname == "jovyan":
        return os.environ.get("JUPYTERHUB_USER", "jovyan")
    else:
        return osname


def find_owner(path):
    """Get the username of the owner of path."""
    if pwd is None:
        raise OSError("find_owner cannot be called on Windows")
    return pwd.getpwuid(os.stat(os.path.abspath(path)).st_uid).pw_name


def self_owned(path):
    """Is the path owned by the current user of this process?"""
    return get_osusername() == find_owner(os.path.abspath(path))


def is_ignored(filename: str, ignore_globs: List[str] = None) -> bool:
    """Determines whether a filename should be ignored, based on whether it
    matches any file glob in the given list. Note that this only matches on the
    base filename itself, not the full path."""
    if ignore_globs is None:
        return False
    dirname = os.path.dirname(filename)
    for expr in ignore_globs:
        globs = glob.glob(os.path.join(dirname, expr))
        if filename in globs:
            return True
    return False


def ignore_patterns(exclude=None, include=None, max_file_size=None, log=None):
    """
    Function that can be used as :func:`shutils.copytree` ignore parameter.

    This is a generalization of :func:`shutils.ignore_patterns` that supports
    include globs, exclude globs, max file size, and logging.

    Parameters
    ----------
    exclude: list or None
        A list of filename globs or None (the default)
    include: list or None
        A list of filename globs or None (the default)
    max_file_size: int or float
        The max file size, in kilobytes
    log: logging.Logger or None (the default)

    Returns
    -------

    A function taking a directory name and list of file/directory
    names and returning the list of file/directory names to be
    ignored.

    A file/directory is ignored as soon as it is either excluded, or
    not included explicitely, or too large.

    If a logger is provided, a warning is logged for files too large
    and a debug message for otherwise ignored files.
    """

    def ignore_patterns(directory, filelist):
        ignored = []
        for filename in filelist:
            fullname = os.path.join(directory, filename)
            if exclude and any(fnmatch.fnmatch(filename, glob) for glob in exclude):
                if log:
                    log.debug(
                        "Ignoring excluded file '{}' (see config option CourseDirectory.ignore)".format(
                            fullname
                        )
                    )
                ignored.append(filename)
            else:
                if os.path.isfile(fullname):
                    if include and not any(fnmatch.fnmatch(filename, glob) for glob in include):
                        if log:
                            log.debug(
                                "Ignoring non included file '{}' (see config option CourseDirectory.include)".format(
                                    fullname
                                )
                            )
                        ignored.append(filename)
                    elif max_file_size and os.path.getsize(fullname) > 1000 * max_file_size:
                        if log:
                            log.warning(
                                "Ignoring file too large '{}' (see config option CourseDirectory.max_file_size)".format(
                                    fullname
                                )
                            )
                        ignored.append(filename)
        return ignored

    return ignore_patterns


def find_all_files(path: str, exclude: List[str] = None) -> List[str]:
    """Recursively finds all filenames rooted at `path`, optionally excluding
    some based on filename globs."""
    files = []
    to_skip = []
    for dirname, dirnames, filenames in os.walk(path):
        if is_ignored(dirname, exclude) or dirname in to_skip:
            to_skip.extend([os.path.join(dirname, x) for x in dirnames])
            continue
        for filename in filenames:
            fullpath = os.path.join(dirname, filename)
            if is_ignored(fullpath, exclude):
                continue
            else:
                files.append(fullpath)
    return files


def find_all_notebooks(path):
    """Return a sorted list of notebooks recursively found rooted at `path`."""
    notebooks = list()
    rootpath = os.path.abspath(path)
    for _file in find_all_files(rootpath):
        if os.path.splitext(_file)[-1] == ".ipynb":
            notebooks.append(os.path.relpath(_file, rootpath))
    notebooks.sort()
    return notebooks


def full_split(path: str) -> Tuple[str, ...]:
    rest, last = os.path.split(path)
    if last == path:
        return (path,)
    elif rest == path:
        return (rest,)
    else:
        return full_split(rest) + (last,)


@contextlib.contextmanager
def chdir(dirname: str) -> Iterator:
    currdir = os.getcwd()
    if dirname:
        os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(currdir)


@contextlib.contextmanager
def setenv(**kwargs: Any) -> Iterator:
    previous_env = {}
    for key, value in kwargs.items():
        previous_env[key] = os.environ.get(value)
        os.environ[key] = value
    yield
    for key, value in kwargs.items():
        if previous_env[key] is None:
            del os.environ[key]
        else:
            os.environ[key] = previous_env[key]


def rmtree(path: str) -> None:
    # for windows, we need to go through and make sure everything
    # is writeable, otherwise rmtree will fail
    if sys.platform == "win32":
        for dirname, _, filenames in os.walk(path):
            os.chmod(dirname, stat.S_IWRITE)
            for filename in filenames:
                os.chmod(os.path.join(dirname, filename), stat.S_IWRITE)

    # now we can remove the path
    shutil.rmtree(path)


def remove(path: str) -> None:
    # for windows, we need to make sure that the file is writeable,
    # otherwise remove will fail
    if sys.platform == "win32":
        os.chmod(path, stat.S_IWRITE)

    # now we can remove the path
    os.remove(path)


def unzip(src, dest, zip_ext=None, create_own_folder=False, tree=False):
    """Extract all content from an archive file to a destination folder.

    Parameters
    ----------
    src: str
        Absolute path to the archive file ('/path/to/archive_filename.zip')
    dest: str
        Asolute path to extract all content to ('/path/to/extract/')
    zip_ext: list
        Valid zip file extensions. Default: ['.zip', '.gz']
    create_own_folder: bool
        Create a sub-folder in 'dest' with the archive file name if True
        ('/path/to/extract/archive_filename/'). Default: False
    tree: bool
        Extract archive files within archive files (into their own
        sub-directory) if True. Default: False
    """
    zip_ext = list(zip_ext or [".zip", ".gz"])
    filename, ext = os.path.splitext(os.path.basename(src))
    if ext not in zip_ext:
        raise ValueError("Invalid archive file extension {}: {}".format(ext, src))
    if not check_directory(dest, write=True, execute=True):
        raise OSError("Directory not found or unwritable: {}".format(dest))

    if create_own_folder:
        # double splitext for .tar.gz
        fname, ext = os.path.splitext(os.path.basename(filename))
        if ext == ".tar":
            filename = fname
        dest = os.path.join(dest, filename)
        if not os.path.isdir(dest):
            os.makedirs(dest)

    safe_unpack_archive(src, dest)

    # extract flat, don't extract archive files within archive files
    if not tree:
        return

    def find_archive_files(skip):
        found = []
        # find archive files in dest that are not in skip
        for root, _, filenames in os.walk(dest):
            for basename in filenames:
                src_file = os.path.join(root, basename)
                _, ext = os.path.splitext(basename)
                if ext in zip_ext and src_file not in skip:
                    found.append(src_file)
        return found

    skip = []
    new_files = find_archive_files(skip)
    # keep walking dest until no new archive files are found
    while new_files:
        # unzip (flat) new archive files found in dest
        for src_file in new_files:
            dest_path = os.path.split(src_file)[0]
            unzip(src_file, dest_path, zip_ext=zip_ext, create_own_folder=True, tree=False)
            skip.append(src_file)
        new_files = find_archive_files(skip)


@contextlib.contextmanager
def temp_attrs(app, **newvals):
    oldvals = {}
    for k, v in newvals.items():
        oldvals[k] = getattr(app, k)
        setattr(app, k, v)

    yield app

    for k, v in oldvals.items():
        setattr(app, k, v)


def capture_log(app, fmt="[%(levelname)s] %(message)s"):
    """Adds an extra handler to the given application the logs to a string
    buffer, calls ``app.start()``, and returns the log output. The extra
    handler is removed from the application before returning.

    Parameters
    ----------
    app: LoggingConfigurable
        An application, withh the `.start()` method implemented
    fmt: string
        A format string for formatting log messages

    Returns
    -------
    A dictionary with the following keys (error and log may or may not be present):

        - success (bool): whether or not the operation completed successfully
        - error (string): formatted traceback
        - log (string): captured log output

    """
    log_buff = io.StringIO()
    handler = logging.StreamHandler(log_buff)
    formatter = LogFormatter(fmt="[%(levelname)s] %(message)s")
    handler.setFormatter(formatter)
    app.log.addHandler(handler)

    try:
        app.start()

    except Exception:
        log_buff.flush()
        val = log_buff.getvalue()
        result = {"success": False}
        result["error"] = traceback.format_exc()
        if val:
            result["log"] = val

    else:
        log_buff.flush()
        val = log_buff.getvalue()
        result = {"success": True}
        if val:
            result["log"] = val

    finally:
        log_buff.close()
        app.log.removeHandler(handler)

    return result


def notebook_hash(path, unique_key=None):
    m = hashlib.md5()
    m.update(open(path, "rb").read())
    if unique_key:
        m.update(to_bytes(unique_key))
    return m.hexdigest()


def make_unique_key(course_id, assignment_id, notebook_id, student_id, timestamp):
    return "+".join([course_id, assignment_id, notebook_id, student_id, timestamp])



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
import sqlite3

from sqlalchemy import Engine, event
from sqlalchemy.orm import declarative_base

from grader_service.api.models.base_model import Model

Base = declarative_base()


class Serializable(object):
    @property
    def model(self) -> Model:
        return Model()

    def serialize(self) -> dict:
        return self.model.to_dict()


class DeleteState(enum.IntEnum):
    active = 0
    deleted = 1


# Register an event listener that enables foreign key constraints for SQLite database.
# The code was copied from the SQLAlchemy documentation:
# https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """
    Enable foreign key constraints for SQLite connections.

    This function is a no-op if the database is not sqlite.
    """
    if isinstance(dbapi_connection, sqlite3.Connection):
        # the sqlite3 driver will not set PRAGMA foreign_keys
        # if autocommit=False; set to True temporarily
        ac = dbapi_connection.autocommit
        dbapi_connection.autocommit = True

        cursor = dbapi_connection.cursor()
        # Note: this is a SQLite-specific pragma
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()

        # restore previous autocommit setting
        dbapi_connection.autocommit = ac



"""Use defined enums as values of submission's statuses

Revision ID: 597857864aed
Revises: f1ae66d52ad9
Create Date: 2025-08-05 17:39:45.007718

"""

from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "597857864aed"
down_revision = "f1ae66d52ad9"
branch_labels = None
depends_on = None


# Enum values
AUTO_STATUS_OLD = ["not_graded", "pending", "automatically_graded", "grading_failed"]
AUTO_STATUS_NEW = [v.upper() for v in AUTO_STATUS_OLD]

MANUAL_STATUS_OLD = ["not_graded", "manually_graded", "being_edited"]
MANUAL_STATUS_NEW = [v.upper() for v in MANUAL_STATUS_OLD]

FEEDBACK_STATUS_OLD = [
    "not_generated",
    "generating",
    "generated",
    "generation_failed",
    "feedback_outdated",
]
FEEDBACK_STATUS_NEW = [v.upper() for v in FEEDBACK_STATUS_OLD]


def upgrade():
    bind = op.get_bind()
    dialect = bind.dialect.name

    if dialect == "postgresql":
        # PostgreSQL: create new enums and swap
        _upgrade_postgresql(bind)
    else:
        # SQLite (or others): just update text
        _upgrade_sqlite()


def downgrade():
    bind = op.get_bind()
    dialect = bind.dialect.name

    if dialect == "postgresql":
        _downgrade_postgresql(bind)
    else:
        _downgrade_sqlite()


# ----------------------
# PostgreSQL helpers
# ----------------------
def _upgrade_postgresql(bind):
    # AUTO_STATUS
    auto_status_new = postgresql.ENUM(*AUTO_STATUS_NEW, name="auto_status_new")
    auto_status_new.create(bind, checkfirst=False)
    op.alter_column(
        "submission",
        "auto_status",
        type_=auto_status_new,
        postgresql_using="UPPER(auto_status::text)::auto_status_new",
    )
    op.execute("DROP TYPE auto_status")
    op.execute("ALTER TYPE auto_status_new RENAME TO auto_status")

    # MANUAL_STATUS
    manual_status_new = postgresql.ENUM(*MANUAL_STATUS_NEW, name="manual_status_new")
    manual_status_new.create(bind, checkfirst=False)
    op.alter_column(
        "submission",
        "manual_status",
        type_=manual_status_new,
        postgresql_using="UPPER(manual_status::text)::manual_status_new",
    )
    op.execute("DROP TYPE manual_status")
    op.execute("ALTER TYPE manual_status_new RENAME TO manual_status")

    # FEEDBACK_STATUS
    feedback_status_new = postgresql.ENUM(*FEEDBACK_STATUS_NEW, name="feedback_status_new")
    feedback_status_new.create(bind, checkfirst=False)

    # Drop default first
    op.execute("ALTER TABLE submission ALTER COLUMN feedback_status DROP DEFAULT")

    op.alter_column(
        "submission",
        "feedback_status",
        type_=feedback_status_new,
        postgresql_using="UPPER(feedback_status::text)::feedback_status_new",
    )
    op.execute("DROP TYPE feedback_status")
    op.execute("ALTER TYPE feedback_status_new RENAME TO feedback_status")


def _downgrade_postgresql(bind):
    auto_status_old = postgresql.ENUM(*AUTO_STATUS_OLD, name="auto_status_old")
    auto_status_old.create(bind, checkfirst=False)
    op.alter_column(
        "submission",
        "auto_status",
        type_=auto_status_old,
        postgresql_using="LOWER(auto_status::text)::auto_status_old",
    )
    op.execute("DROP TYPE auto_status")
    op.execute("ALTER TYPE auto_status_old RENAME TO auto_status")

    manual_status_old = postgresql.ENUM(*MANUAL_STATUS_OLD, name="manual_status_old")
    manual_status_old.create(bind, checkfirst=False)
    op.alter_column(
        "submission",
        "manual_status",
        type_=manual_status_old,
        postgresql_using="LOWER(manual_status::text)::manual_status_old",
    )
    op.execute("DROP TYPE manual_status")
    op.execute("ALTER TYPE manual_status_old RENAME TO manual_status")

    feedback_status_old = postgresql.ENUM(*FEEDBACK_STATUS_OLD, name="feedback_status_old")
    feedback_status_old.create(bind, checkfirst=False)
    op.alter_column(
        "submission",
        "feedback_status",
        type_=feedback_status_old,
        postgresql_using="LOWER(feedback_status::text)::feedback_status_old",
    )
    op.execute("DROP TYPE feedback_status")
    op.execute("ALTER TYPE feedback_status_old RENAME TO feedback_status")
    # restore default
    op.execute("ALTER TABLE submission ALTER COLUMN feedback_status SET DEFAULT 'not_generated'")


# ----------------------
# SQLite helpers
# ----------------------
def _upgrade_sqlite():
    op.execute("UPDATE submission SET auto_status = UPPER(auto_status)")
    op.execute("UPDATE submission SET manual_status = UPPER(manual_status)")
    op.execute("UPDATE submission SET feedback_status = UPPER(feedback_status)")


def _downgrade_sqlite():
    op.execute("UPDATE submission SET auto_status = LOWER(auto_status)")
    op.execute("UPDATE submission SET manual_status = LOWER(manual_status)")
    op.execute("UPDATE submission SET feedback_status = LOWER(feedback_status)")



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import os
import shlex
import subprocess
from pathlib import Path
from string import Template
from typing import List, Optional

from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from tornado.ioloop import IOLoop
from tornado.iostream import StreamClosedError
from tornado.process import Subprocess
from tornado.web import HTTPError, stream_request_body

from grader_service.handlers.base_handler import GraderBaseHandler, RequestHandlerConfig
from grader_service.handlers.handler_utils import GitRepoType
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import Submission
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler


class GitBaseHandler(GraderBaseHandler):
    async def data_received(self, chunk: bytes):
        return self.process.stdin.write(chunk)

    def write_error(self, status_code: int, **kwargs) -> None:
        self.clear()
        if status_code == 401:
            self.set_header("WWW-Authenticate", 'Basic realm="User Visible Realm"')
        self.set_status(status_code)

    def on_finish(self):
        if hasattr(
            self, "process"
        ):  # if we exit super prepare (authentication) process is not created
            if self.process.stdin is not None:
                self.process.stdin.close()
            if self.process.stdout is not None:
                self.process.stdout.close()
            if self.process.stderr is not None:
                self.process.stderr.close()
            IOLoop.current().spawn_callback(self.process.wait_for_exit)

    async def git_response(self):
        try:
            while data := await self.process.stdout.read_bytes(8192, partial=True):
                self.write(data)
                await self.flush()
        except StreamClosedError:
            pass
        except Exception as e:
            self.log.error(f"Error from git response {e}")
            raise HTTPError(500, str(e))

    def _check_git_repo_permissions(self, rpc: str, role: Role, pathlets: List[str]):
        repo_type: str = pathlets[2]

        if role.role == Scope.student:
            # 1. no source or release interaction with source repo for students
            # 2. no pull allowed for autograde for students
            if (repo_type in {GitRepoType.SOURCE, GitRepoType.RELEASE, GitRepoType.EDIT}) or (
                repo_type == GitRepoType.AUTOGRADE and rpc == "upload-pack"
            ):
                raise HTTPError(403)

            # 3. students should not be able to pull other submissions
            #    -> add query param for sub_id
            if (repo_type == GitRepoType.FEEDBACK) and (rpc == "upload-pack"):
                try:
                    sub_id = int(pathlets[3])
                except (ValueError, IndexError):
                    raise HTTPError(403)
                submission = self.session.query(Submission).get(sub_id)
                if submission is None or submission.user_id != self.user.id:
                    raise HTTPError(403)

        # 4. no push allowed for autograde and feedback
        #    -> the autograder executor can push locally (will bypass this)
        if (repo_type in {GitRepoType.AUTOGRADE, GitRepoType.FEEDBACK}) and (
            rpc in ["send-pack", "receive-pack"]
        ):
            raise HTTPError(403)

    def gitlookup(self, rpc: str) -> Optional[str]:
        pathlets = self.request.path.strip("/").split("/")
        # check if request is sent using jupyterhub as a proxy
        # if yes, remove services/grader path prefix
        assert len(pathlets) > 0
        if pathlets[0] == "services":
            pathlets = pathlets[2:]

        # pathlets should look like this
        # pathlets = ['git',
        #             'lecture_code', 'assignment_id', 'repo_type', ...]
        if len(pathlets) < 4:
            return None

        # cut git prefix
        pathlets = pathlets[1:]
        lect_code, assign_id, repo_type, *pathlets_tail = pathlets

        # Repo type "assignment" has been replaced by "user", so this should not happen,
        # but we are leaving this check for the time being, just to be on the safe side:
        if repo_type == "assignment":
            self.log.warning("Deprecated repo_type: 'assignment'! Setting it to 'user'")
            repo_type = GitRepoType.USER

        try:
            repo_type = GitRepoType(repo_type)
        except ValueError:
            return None

        # get lecture and assignment if they exist
        try:
            lecture = self.session.query(Lecture).filter(Lecture.code == lect_code).one()
        except NoResultFound:
            raise HTTPError(404, reason="Lecture was not found")
        except MultipleResultsFound:
            raise HTTPError(500, reason="Found more than one lecture")

        role = self.session.get(Role, (self.user.id, lecture.id))
        self._check_git_repo_permissions(rpc, role, pathlets)

        try:
            assignment = self.get_assignment(lecture.id, int(assign_id))
        except ValueError:
            raise HTTPError(404, reason="Assignment not found")

        # create directories once we know they exist in the database
        lecture_path = os.path.abspath(os.path.join(self.gitbase, lect_code))
        assignment_path = os.path.abspath(os.path.join(lecture_path, assign_id))
        if not os.path.exists(lecture_path):
            os.mkdir(lecture_path)
        if not os.path.exists(assignment_path):
            os.mkdir(assignment_path)

        submission = None
        if repo_type in {GitRepoType.AUTOGRADE, GitRepoType.FEEDBACK, GitRepoType.EDIT}:
            try:
                sub_id = int(pathlets_tail[0])
            except (ValueError, IndexError):
                raise HTTPError(403, "Invalid or missing submission id")
            submission = self.get_submission(lecture.id, assignment.id, sub_id)

        path = self.construct_git_dir(repo_type, lecture, assignment, submission=submission)
        if path is None:
            return None

        os.makedirs(os.path.dirname(path), exist_ok=True)
        is_git = self.is_base_git_dir(path)
        # return git repo
        if os.path.exists(path) and is_git:
            self.write_pre_receive_hook(path)
            return path
        else:
            os.mkdir(path)
            # this path has to be a git dir -> call git init
            try:
                self.log.info("Running: git init --bare")
                subprocess.run(["git", "init", "--bare", path], check=True)
            except subprocess.CalledProcessError:
                return None

            if repo_type == GitRepoType.USER:
                repo_path_release = self.construct_git_dir(
                    GitRepoType.RELEASE, assignment.lecture, assignment
                )
                if not os.path.exists(repo_path_release):
                    return None
                self.duplicate_release_repo(
                    repo_path_release=repo_path_release,
                    repo_path_user=path,
                    assignment=assignment,
                    message="Initialize with Release",
                    checkout_main=True,
                )

            self.write_pre_receive_hook(path)
            return path

    def write_pre_receive_hook(self, path: str):
        hook_dir = os.path.join(path, "hooks")
        if not os.path.exists(hook_dir):
            os.mkdir(hook_dir)

        hook_file = os.path.join(hook_dir, "pre-receive")
        if not os.path.exists(hook_file):
            tpl = Template(self._read_hook_template())
            hook = tpl.safe_substitute(
                {
                    "tpl_max_file_size": self._get_hook_max_file_size(),
                    "tpl_file_extensions": self._get_hook_file_allow_pattern(),
                    "tpl_max_file_count": self._get_hook_max_file_count(),
                }
            )
            with open(hook_file, "wt") as f:
                os.chmod(hook_file, 0o755)
                f.write(hook)

    @staticmethod
    def _get_hook_file_allow_pattern(extensions: Optional[List[str]] = None) -> str:
        pattern = ""
        if extensions is None:
            req_handler_conf = RequestHandlerConfig.instance()
            extensions = req_handler_conf.git_allowed_file_extensions
        if len(extensions) > 0:
            allow_patterns = ["\\." + s.strip(".").replace(".", "\\.") for s in extensions]
            pattern = "|".join(allow_patterns)
        return pattern

    @staticmethod
    def _get_hook_max_file_size():
        return RequestHandlerConfig.instance().git_max_file_size_mb

    @staticmethod
    def _get_hook_max_file_count():
        return RequestHandlerConfig.instance().git_max_file_count

    @staticmethod
    def _read_hook_template() -> str:
        file_path = Path(__file__).parent / "hook_templates" / "pre-receive"
        with open(file_path, mode="rt") as f:
            return f.read()

    @staticmethod
    def _create_path(path):
        if not os.path.exists(path):
            os.mkdir(path)

    def get_gitdir(self, rpc: str):
        """Determine the git repository for this request"""
        gitdir = self.gitlookup(rpc)
        if gitdir is None:
            raise HTTPError(404, reason="unable to find repository")
        self.log.info("Accessing git at: %s", gitdir)

        return gitdir


@register_handler(path="/.*/git-(.*)", version_specifier=VersionSpecifier.NONE)
@stream_request_body
class RPCHandler(GitBaseHandler):
    """Request handler for RPC calls

    Use this handler to handle example.git/git-upload-pack
    and example.git/git-receive-pack URLs"""

    async def prepare(self):
        await super().prepare()
        self.rpc = self.path_args[0]
        self.gitdir = self.get_gitdir(rpc=self.rpc)
        self.cmd = f'git {self.rpc} --stateless-rpc "{self.gitdir}"'
        self.log.info(f"Running command: {self.cmd}")
        self.process = Subprocess(
            shlex.split(self.cmd),
            stdin=Subprocess.STREAM,
            stderr=Subprocess.STREAM,
            stdout=Subprocess.STREAM,
        )

    async def post(self, rpc):
        self.set_header("Content-Type", "application/x-git-%s-result" % rpc)
        self.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")
        await self.git_response()
        await self.finish()


@register_handler(path="/.*/info/refs", version_specifier=VersionSpecifier.NONE)
class InfoRefsHandler(GitBaseHandler):
    """Request handler for info/refs

    Use this handler to handle example.git/info/refs?service= URLs"""

    async def prepare(self):
        await super().prepare()
        if self.get_status() != 200:
            return
        self.rpc = self.get_argument("service")[4:]
        self.cmd = f'git {self.rpc} --stateless-rpc --advertise-refs "{self.get_gitdir(self.rpc)}"'
        self.log.info(f"Running command: {self.cmd}")
        self.process = Subprocess(
            shlex.split(self.cmd),
            stdin=Subprocess.STREAM,
            stderr=Subprocess.STREAM,
            stdout=Subprocess.STREAM,
        )

    async def get(self):
        self.set_header("Content-Type", "application/x-git-%s-advertisement" % self.rpc)
        self.set_header("Cache-Control", "no-store, no-cache, must-revalidate, max-age=0")

        prelude = f"# service=git-{self.rpc}\n0000"
        size = str(hex(len(prelude))[2:].rjust(4, "0"))
        self.write(size)
        self.write(prelude)
        await self.flush()

        await self.git_response()
        await self.finish()
